@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2011-2014 Canonical
+ * Copyright (C) 2011-2015 Canonical
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -15,6 +15,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
+ * Author: Colin Ian King <colin.king@canonical.com>
  */
 
 #include <stdlib.h>
@@ -44,33 +45,33 @@
 #include <linux/netlink.h>
 #include <linux/cn_proc.h>
 
-#define APP_NAME		"powerstat"
 #define MIN_RUN_DURATION	(5*60)		/* We recommend a run of 5 minutes */
-#define SAMPLE_DELAY		(10)		/* Delay between samples in seconds */
+#define SAMPLE_DELAY		(10.0)		/* Delay between samples in seconds */
 #define ROLLING_AVERAGE_SECS	(120)		/* 2 minute rolling average for power usage calculation */
 #define STANDARD_AVERAGE_SECS	(120)
 #define MAX_MEASUREMENTS 	(ROLLING_AVERAGE_SECS + 10)
 #define START_DELAY		(3*60)		/* Delay to wait before sampling */
 #define MAX_PIDS		(32769)		/* Hash Max PIDs */
 #define	RATE_ZERO_LIMIT		(0.001)		/* Less than this we call the power rate zero */
-#define IDLE_THRESHOLD		(98)		/* Less than this and we assume the machine is not idle */
+#define IDLE_THRESHOLD		(98)		/* Less than this and we assume the device is not idle */
 
-#define CPU_USER	0
-#define CPU_NICE	1
-#define CPU_SYS		2
-#define CPU_IDLE	3
-#define CPU_IOWAIT	4
-#define CPU_IRQ		5
-#define CPU_SOFTIRQ	6
-#define CPU_INTR	7
-#define CPU_CTXT	8
-#define CPU_PROCS_RUN	9
-#define CPU_PROCS_BLK	10
-#define POWER_RATE	11
-#define PROC_FORK	12
-#define PROC_EXEC	13
-#define PROC_EXIT	14
-#define MAX_VALUES	15
+/* Statistics gathered from /proc/stat and process activity */
+#define CPU_USER		(0)
+#define CPU_NICE		(1)
+#define CPU_SYS			(2)
+#define CPU_IDLE		(3)
+#define CPU_IOWAIT		(4)
+#define CPU_IRQ			(5)
+#define CPU_SOFTIRQ		(6)
+#define CPU_INTR		(7)
+#define CPU_CTXT		(8)
+#define CPU_PROCS_RUN		(9)
+#define CPU_PROCS_BLK		(10)
+#define POWER_RATE		(11)
+#define PROC_FORK		(12)
+#define PROC_EXEC		(13)
+#define PROC_EXIT		(14)
+#define MAX_VALUES		(15)
 
 /* Arg opt flags */
 #define OPTS_SHOW_PROC_ACTIVITY	(0x0001)	/* dump out process activity */
@@ -96,8 +97,8 @@
 
 /* Measurement entry */
 typedef struct {
-	double	value;	/* Measurment value */
-	time_t	when;	/* When it was measured */
+	double	value;		/* Measurement value */
+	time_t	when;		/* When it was measured */
 } measurement_t;
 
 /* Statistics entry */
@@ -126,14 +127,73 @@ typedef struct {
 
 static proc_info_t *proc_info[MAX_PIDS];	/* Proc hash table */
 static long int max_readings;			/* number of samples to gather */
-static long int sample_delay   = SAMPLE_DELAY;	/* time between each sample in secs */
+static double sample_delay = SAMPLE_DELAY;	/* time between each sample in secs */
 static long int start_delay = START_DELAY;	/* seconds before we start displaying stats */
 static double idle_threshold = IDLE_THRESHOLD;	/* lower than this and the CPU is busy */
 static log_t infolog;				/* log */
 static int opts;				/* opt arg opt flags */
 static volatile int stop_recv;			/* sighandler stop flag */
 static bool power_calc_from_capacity = false;	/* true of power is calculated via capacity change */
+static const char *app_name = "powerstat";	/* name of application */
 
+/*
+ *  Attempt to catch a range of signals so
+ *  we can clean
+ */
+static const int signals[] = {
+	/* POSIX.1-1990 */
+#ifdef SIGHUP
+	SIGHUP,
+#endif
+#ifdef SIGINT
+	SIGINT,
+#endif
+#ifdef SIGQUIT
+	SIGQUIT,
+#endif
+#ifdef SIGILL
+	SIGILL,
+#endif
+#ifdef SIGABRT
+	SIGABRT,
+#endif
+#ifdef SIGFPE
+	SIGFPE,
+#endif
+#ifdef SIGTERM
+	SIGTERM,
+#endif
+#ifdef SIGUSR1
+	SIGUSR1,
+#endif
+#ifdef SIGUSR2
+	SIGUSR2,
+	/* POSIX.1-2001 */
+#endif
+#ifdef SIGXCPU
+	SIGXCPU,
+#endif
+#ifdef SIGXFSZ
+	SIGXFSZ,
+#endif
+	/* Linux various */
+#ifdef SIGIOT
+	SIGIOT,
+#endif
+#ifdef SIGSTKFLT
+	SIGSTKFLT,
+#endif
+#ifdef SIGPWR
+	SIGPWR,
+#endif
+#ifdef SIGINFO
+	SIGINFO,
+#endif
+#ifdef SIGVTALRM
+	SIGVTALRM,
+#endif
+	-1,
+};
 
 /*
  *  file_get()
@@ -148,7 +208,7 @@ static char *file_get(const char *const file)
 		return NULL;
 
 	if (fgets(buffer, sizeof(buffer), fp) == NULL) {
-		fclose(fp);
+		(void)fclose(fp);
 		return NULL;
 	}
 
@@ -177,19 +237,65 @@ static int tty_height(void)
 	return 25;	/* else standard tty 80x25 */
 }
 
+
 /*
- *  time_now()
+ *  timeval_to_double
+ *	timeval to a double (in seconds)
+ */
+static inline double timeval_to_double(const struct timeval *const tv)
+{
+	return (double)tv->tv_sec + ((double)tv->tv_usec / 1000000.0);
+}
+
+
+/*
+ *  double_to_timeval
+ *	seconds in double to timeval
+ */
+static inline struct timeval double_to_timeval(const double val)
+{
+	struct timeval tv;
+
+	tv.tv_sec = val;
+	tv.tv_usec = (val - (time_t)val) * 1000000.0;
+
+	return tv;
+}
+
+/*
+ *  gettime_to_double()
+ *	get time as a double
+ */
+static double gettime_to_double(void)
+{
+	struct timeval tv;
+
+        if (gettimeofday(&tv, NULL) < 0) {
+                fprintf(stderr, "gettimeofday failed: errno=%d (%s).\n",
+                        errno, strerror(errno));
+		return -1.0;
+        }
+        return timeval_to_double(&tv);
+}
+
+/*
+ *  get_time()
  *	Gather current time in buffer
  */
-static void time_now(char *const buffer, const size_t buflen)
+static void get_time(char *const buffer, const size_t buflen)
 {
 	struct tm tm;
 	time_t now;
 
-	(void)time(&now);
+	now = time(NULL);
+	if (now == ((time_t) -1)) {
+		/* Unknown time! */
+		(void)snprintf(buffer, buflen, "--:--:-- ");
+		return;
+	}
 	(void)localtime_r(&now, &tm);
 
-	snprintf(buffer, buflen, "%2.2d:%2.2d:%2.2d ",
+	(void)snprintf(buffer, buflen, "%2.2d:%2.2d:%2.2d ",
 		tm.tm_hour, tm.tm_min, tm.tm_sec);
 }
 
@@ -216,17 +322,21 @@ static int log_printf(const char *const fmt, ...)
 	size_t len;
 
 	va_start(ap, fmt);
-	time_now(tmbuffer, sizeof(tmbuffer));
-	vsnprintf(buffer, sizeof(buffer), fmt, ap);
+	get_time(tmbuffer, sizeof(tmbuffer));
+	(void)vsnprintf(buffer, sizeof(buffer), fmt, ap);
 	va_end(ap);
 
 	if ((log_item = calloc(1, sizeof(log_t))) == NULL) {
-		fprintf(stderr, "Out of memory allocating log item\n");
+		fprintf(stderr, "Out of memory allocating log item.\n");
 		return -1;
 	}
 	len = strlen(buffer) + strlen(tmbuffer) + 1;
-	log_item->text = calloc(1, len);
-	snprintf(log_item->text, len, "%s%s", tmbuffer, buffer);
+	if ((log_item->text = calloc(1, len)) == NULL) {
+		free(log_item);
+		fprintf(stderr, "Out of memory allocating log item text.\n");
+		return -1;
+	}
+	(void)snprintf(log_item->text, len, "%s%s", tmbuffer, buffer);
 
 	if (infolog.head == NULL)
 		infolog.head = log_item;
@@ -249,7 +359,7 @@ static void log_dump(void)
 	if (infolog.head != NULL)
 		printf("\nLog of fork()/exec()/exit() calls:\n");
 
-	for (log_item = infolog.head; log_item != NULL; log_item = log_item->next)
+	for (log_item = infolog.head; log_item; log_item = log_item->next)
 		printf("%s", log_item->text);
 }
 
@@ -261,7 +371,7 @@ static void log_free(void)
 {
 	log_item_t *log_item = infolog.head;
 
-	while (log_item != NULL) {
+	while (log_item) {
 		log_item_t *log_next = log_item->next;
 		free(log_item->text);
 		free(log_item);
@@ -273,10 +383,10 @@ static void log_free(void)
 }
 
 /*
- *  handle_sigint()
- *	catch SIGINT and flag a stop
+ *  handle_sig()
+ *	catch signals and flag a stop
  */
-static void handle_sigint(int dummy)
+static void handle_sig(int dummy)
 {
 	(void)dummy;
 	stop_recv = 1;
@@ -294,7 +404,8 @@ static int netlink_connect(void)
     	if ((sock = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_CONNECTOR)) < 0) {
 		if (errno == EPROTONOSUPPORT)
 			return -EPROTONOSUPPORT;
-		fprintf(stderr, "Socket failed: %s\n", strerror(errno));
+		fprintf(stderr, "socket failed: errno=%d (%s).\n",
+			errno, strerror(errno));
 		return -1;
 	}
 
@@ -304,7 +415,8 @@ static int netlink_connect(void)
 	addr.nl_groups = CN_IDX_PROC;
 
 	if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
-		fprintf(stderr, "Bind failed: %s\n", strerror(errno));
+		fprintf(stderr, "Bind failed: errno=%d (%s).\n",
+			errno, strerror(errno));
 		(void)close(sock);
 		return -1;
 	}
@@ -345,19 +457,31 @@ static int netlink_listen(const int sock)
 }
 
 /*
- *  stats_clear()
- *	clear stats
+ *  stats_set()
+ *	set stats
  */
-static void stats_clear(stats_t *const stats)
+static void stats_set(
+	stats_t *const stats,
+	const double value,
+	const bool inaccurate)
 {
 	int i;
 
 	for (i = 0; i < MAX_VALUES; i++) {
-		stats->value[i] = 0.0;
-		stats->inaccurate[i] = false;
+		stats->value[i] = value;
+		stats->inaccurate[i] = inaccurate;
 	}
 }
 
+/*
+ *  stats_set()
+ *	clear stats
+ */
+static inline void stats_clear(stats_t *const stats)
+{
+	stats_set(stats, 0.0, false);
+}
+
 /*
  *  stats_clear_all()
  *	zero stats data
@@ -374,32 +498,59 @@ static void stats_clear_all(stats_t *const stats, const long int n)
  *  stats_read()
  *	gather pertinent /proc/stat data
  */
-static int stats_read(stats_t *const info)
+static int stats_read(stats_t *const stats)
 {
 	FILE *fp;
 	char buf[4096];
+	int i, j;
 
-	if ((fp = fopen("/proc/stat", "r")) == NULL)
+	static int indices[] = {
+		CPU_USER, CPU_NICE, CPU_SYS, CPU_IDLE,
+		CPU_IOWAIT, CPU_IRQ, CPU_SOFTIRQ, CPU_CTXT,
+		CPU_INTR, CPU_PROCS_RUN, CPU_PROCS_BLK, -1
+	};
+
+	for (i = 0; (j = indices[i]) != -1; i++) {
+		stats->value[j] = 0.0;
+		stats->inaccurate[j] = true;
+	}
+
+	if ((fp = fopen("/proc/stat", "r")) == NULL) {
+		fprintf(stderr, "Cannot read /proc/stat, errno=%d (%s).\n",
+			errno, strerror(errno));
 		return -1;
+	}
 
 	while (fgets(buf, sizeof(buf), fp) != NULL) {
 		if (strncmp(buf, "cpu ", 4) == 0)
-			sscanf(buf, "%*s %15lf %15lf %15lf %15lf %15lf %15lf %15lf",
-				&(info->value[CPU_USER]),
-				&(info->value[CPU_NICE]),
-				&(info->value[CPU_SYS]),
-				&(info->value[CPU_IDLE]),
-				&(info->value[CPU_IOWAIT]),
-				&(info->value[CPU_IRQ]),
-				&(info->value[CPU_SOFTIRQ]));
+			if (sscanf(buf, "%*s %15lf %15lf %15lf %15lf %15lf %15lf %15lf",
+			    &(stats->value[CPU_USER]),
+			    &(stats->value[CPU_NICE]),
+			    &(stats->value[CPU_SYS]),
+			    &(stats->value[CPU_IDLE]),
+			    &(stats->value[CPU_IOWAIT]),
+			    &(stats->value[CPU_IRQ]),
+			    &(stats->value[CPU_SOFTIRQ])) == 7) {
+				stats->inaccurate[CPU_USER] = false;
+				stats->inaccurate[CPU_NICE] = false;
+				stats->inaccurate[CPU_SYS] = false;
+				stats->inaccurate[CPU_IDLE] = false;
+				stats->inaccurate[CPU_IOWAIT] = false;
+				stats->inaccurate[CPU_IRQ] = false;
+				stats->inaccurate[CPU_SOFTIRQ] = false;
+			}
 		if (strncmp(buf, "ctxt ", 5) == 0)
-			sscanf(buf, "%*s %15lf", &(info->value[CPU_CTXT]));
+			if (sscanf(buf, "%*s %15lf", &(stats->value[CPU_CTXT])) == 1)
+				stats->inaccurate[CPU_CTXT] = false;
 		if (strncmp(buf, "intr ", 5) == 0)
-			sscanf(buf, "%*s %15lf", &(info->value[CPU_INTR]));
+			if (sscanf(buf, "%*s %15lf", &(stats->value[CPU_INTR])) == 1)
+				stats->inaccurate[CPU_INTR] = false;
 		if (strncmp(buf, "procs_running ", 14) == 0)
-			sscanf(buf, "%*s %15lf", &(info->value[CPU_PROCS_RUN]));
+			if (sscanf(buf, "%*s %15lf", &(stats->value[CPU_PROCS_RUN])) == 1)
+				stats->inaccurate[CPU_PROCS_RUN] = false;
 		if (strncmp(buf, "procs_blocked ", 14) == 0)
-			sscanf(buf, "%*s %15lf", &(info->value[CPU_PROCS_BLK]));
+			if (sscanf(buf, "%*s %15lf", &(stats->value[CPU_PROCS_BLK])) == 1)
+				stats->inaccurate[CPU_PROCS_BLK] = false;
 	}
 	(void)fclose(fp);
 
@@ -407,10 +558,32 @@ static int stats_read(stats_t *const info)
 }
 
 /*
- *  On Nexus 4 we occasionally get idle time going backwards so
- *  work around this by ensuring we don't get -ve deltas.
+ *  stats_sane()
+ *	check if stats are accurate and calculate a
+ *	sane -ve delta
  */
-#define SANE_STATS(s1, s2) ((s2) - (s1)) < 0.0 ? 0.0 : ((s2) - (s1))
+static double stats_sane(
+	const stats_t *const s1,
+	const stats_t *const s2,
+	const int index)
+{
+	double ret;
+
+	/* Discard inaccurate or empty stats */
+	if (s1->inaccurate[index] || s2->inaccurate[index])
+		return 0.0;
+
+	/*
+	 *  On Nexus 4 we occasionally get idle time going backwards so
+	 *  work around this by ensuring we don't get -ve deltas.
+	 */
+	ret = s2->value[index] - s1->value[index];
+	return ret < 0.0 ? 0.0 : ret;
+}
+
+#define INACCURATE(s1, s2, index)	\
+	(s1->inaccurate[index] | s2->inaccurate[index])
+
 
 /*
  *  stats_gather()
@@ -423,38 +596,48 @@ static bool stats_gather(
 	stats_t *const res)
 {
 	double total;
+	int i, j;
+	bool inaccurate = false;
 
-	res->value[CPU_USER]    = SANE_STATS(s1->value[CPU_USER]   , s2->value[CPU_USER]);
-	res->value[CPU_NICE]    = SANE_STATS(s1->value[CPU_NICE]   , s2->value[CPU_NICE]);
-	res->value[CPU_SYS]     = SANE_STATS(s1->value[CPU_SYS]    , s2->value[CPU_SYS]);
-	res->value[CPU_IDLE]    = SANE_STATS(s1->value[CPU_IDLE]   , s2->value[CPU_IDLE]);
-	res->value[CPU_IOWAIT]  = SANE_STATS(s1->value[CPU_IOWAIT] , s2->value[CPU_IOWAIT]);
-	res->value[CPU_IRQ]     = SANE_STATS(s1->value[CPU_IRQ]    , s2->value[CPU_IRQ]);
-	res->value[CPU_SOFTIRQ] = SANE_STATS(s1->value[CPU_SOFTIRQ], s2->value[CPU_SOFTIRQ]);
-	res->value[CPU_CTXT]	= SANE_STATS(s1->value[CPU_CTXT]   , s2->value[CPU_CTXT]);
-	res->value[CPU_INTR]	= SANE_STATS(s1->value[CPU_INTR]   , s2->value[CPU_INTR]);
+	static int indices[] = {
+		CPU_USER, CPU_NICE, CPU_SYS, CPU_IDLE,
+		CPU_IOWAIT, -1
+	};
+
+	res->value[CPU_USER]    = stats_sane(s1, s2, CPU_USER);
+	res->value[CPU_NICE]    = stats_sane(s1, s2, CPU_NICE);
+	res->value[CPU_SYS]     = stats_sane(s1, s2, CPU_SYS);
+	res->value[CPU_IDLE]    = stats_sane(s1, s2, CPU_IDLE);
+	res->value[CPU_IOWAIT]  = stats_sane(s1, s2, CPU_IOWAIT);
+	res->value[CPU_IRQ]     = stats_sane(s1, s2, CPU_IRQ);
+	res->value[CPU_SOFTIRQ] = stats_sane(s1, s2, CPU_SOFTIRQ);
+	res->value[CPU_CTXT]	= stats_sane(s1, s2, CPU_CTXT);
+	res->value[CPU_INTR]	= stats_sane(s1, s2, CPU_INTR);
+
+	for (i = 0; (j = indices[i]) != -1; i++)
+		inaccurate |= (s1->inaccurate[j] | s2->inaccurate[j]);
 
 	total = res->value[CPU_USER] + res->value[CPU_NICE] +
-		res->value[CPU_SYS] + res->value[CPU_IDLE] + res->value[CPU_IOWAIT];
+		res->value[CPU_SYS] + res->value[CPU_IDLE] +
+		res->value[CPU_IOWAIT];
 
 	/*
 	 * This should not happen, but we need to avoid division
-	 * by zero or weird results
+	 * by zero or weird results if the data is deemed valid
 	 */
-	if (total <= 0.0)
+	if (!inaccurate && total <= 0.0)
 		return false;
 
-	res->value[CPU_USER]   = (100.0 * res->value[CPU_USER]) / total;
-	res->value[CPU_NICE]   = (100.0 * res->value[CPU_NICE]) / total;
-	res->value[CPU_SYS]    = (100.0 * res->value[CPU_SYS]) / total;
-	res->value[CPU_IDLE]   = (100.0 * res->value[CPU_IDLE]) / total;
-	res->value[CPU_IOWAIT] = (100.0 * res->value[CPU_IOWAIT]) / total;
-
-	res->value[CPU_CTXT]   = res->value[CPU_CTXT] / sample_delay;
-	res->value[CPU_INTR]   = res->value[CPU_INTR] / sample_delay;
-
-	res->value[CPU_PROCS_RUN] = s2->value[CPU_PROCS_RUN];
-	res->value[CPU_PROCS_BLK] = s2->value[CPU_PROCS_BLK];
+	for (i = 0; (j = indices[i]) != -1; i++) {
+		res->value[j] = (INACCURATE(s1, s2, j) || (total <= 0.0)) ?
+			NAN : (100.0 * res->value[j]) / total;
+	}
+	res->value[CPU_CTXT] = (INACCURATE(s1, s2, CPU_CTXT) || (sample_delay <= 0.0)) ?
+			NAN : res->value[CPU_CTXT] / sample_delay;
+	res->value[CPU_INTR] = (INACCURATE(s1, s2, CPU_INTR) || (sample_delay <= 0.0)) ?
+			NAN : res->value[CPU_INTR] / sample_delay;
+	res->value[CPU_PROCS_RUN] = s2->inaccurate[CPU_PROCS_RUN] ? NAN : s2->value[CPU_PROCS_RUN];
+	res->value[CPU_PROCS_BLK] = s2->inaccurate[CPU_PROCS_BLK] ? NAN : s2->value[CPU_PROCS_BLK];
 
 	return true;
 }
@@ -484,15 +667,15 @@ static void stats_ruler(void)
 }
 
 /*
- *
- *
+ *  row_increment()
+ *	bump row, reset if hit tty height
  */
 static void row_increment(int *const row)
 {
 	int tty_rows = tty_height();
 
 	(*row)++;
-	if ((tty_rows >2) && (*row >= tty_rows)) {
+	if ((tty_rows > 2) && (*row >= tty_rows)) {
 		stats_headings();
 		*row = 2;
 	}
@@ -511,18 +694,22 @@ static void stats_print(
 
 	if (summary) {
 		if (s->inaccurate[POWER_RATE])
-			snprintf(buf, sizeof(buf), "-N/A-");
+			(void)snprintf(buf, sizeof(buf), "-N/A-");
 		else
-			snprintf(buf, sizeof(buf), "%6.2f", s->value[POWER_RATE]);
+			(void)snprintf(buf, sizeof(buf), "%6.2f",
+				s->value[POWER_RATE]);
 	} else {
-		snprintf(buf, sizeof(buf), "%6.2f%s", s->value[POWER_RATE],
+		(void)snprintf(buf, sizeof(buf), "%6.2f%s",
+			s->value[POWER_RATE],
 			s->inaccurate[POWER_RATE] ? "E" : "");
 	}
 
 	if (opts & OPTS_USE_NETLINK) {
 		char *fmt = summary ?
-			"%8.8s %5.1f %5.1f %5.1f %5.1f %5.1f %4.1f %6.1f %6.1f %4.1f %4.1f %4.1f %s\n" :
-			"%8.8s %5.1f %5.1f %5.1f %5.1f %5.1f %4.0f %6.0f %6.0f %4.0f %4.0f %4.0f %s\n";
+			"%8.8s %5.1f %5.1f %5.1f %5.1f %5.1f "
+			"%4.1f %6.1f %6.1f %4.1f %4.1f %4.1f %s\n" :
+			"%8.8s %5.1f %5.1f %5.1f %5.1f %5.1f "
+			"%4.0f %6.0f %6.0f %4.0f %4.0f %4.0f %s\n";
 		printf(fmt,
 			prefix,
 			s->value[CPU_USER], s->value[CPU_NICE],
@@ -533,8 +720,10 @@ static void stats_print(
 			s->value[PROC_EXIT], buf);
 	} else {
 		char *fmt = summary ?
-			"%8.8s %5.1f %5.1f %5.1f %5.1f %5.1f %4.1f %6.1f %6.1f %s\n" :
-			"%8.8s %5.1f %5.1f %5.1f %5.1f %5.1f %4.0f %6.0f %6.0f %s\n";
+			"%8.8s %5.1f %5.1f %5.1f %5.1f %5.1f "
+			"%4.1f %6.1f %6.1f %s\n" :
+			"%8.8s %5.1f %5.1f %5.1f %5.1f %5.1f "
+			"%4.0f %6.0f %6.0f %s\n";
 		printf(fmt,
 			prefix,
 			s->value[CPU_USER], s->value[CPU_NICE],
@@ -581,7 +770,8 @@ static void stats_average_stddev_min_max(
 			total = 0.0;
 			for (i = 0; i < num; i++) {
 				if (!stats[i].inaccurate[j]) {
-					double diff = (double)stats[i].value[j] - average->value[j];
+					double diff = (double)stats[i].value[j]
+						 - average->value[j];
 					diff = diff * diff;
 					total += diff;
 				}
@@ -619,6 +809,11 @@ static void calc_standard_average(
 	double dw;
 
 	time_now = time(NULL);
+	if (time_now == ((time_t) -1)) {
+		*rate = 0.0;
+		*inaccurate = true;
+		return;
+	}
 
 	if (first) {
 		time_start = time_now;
@@ -663,6 +858,11 @@ static void calc_rolling_average(
 	int i, j;
 
 	time_now = time(NULL);
+	if (time_now == ((time_t) -1)) {
+		*rate = 0.0;
+		*inaccurate = true;
+		return;
+	}
 	measurements[index].value = total_capacity;
 	measurements[index].when  = time_now;
 	index = (index + 1) % MAX_MEASUREMENTS;
@@ -701,6 +901,10 @@ static void calc_rolling_average(
 }
 
 
+/*
+ *  calc_from_capacity()
+ *	calculate either using standard or rolling averages
+ */
 static void calc_from_capacity(
 	const double total_capacity,
 	double *const rate,
@@ -732,7 +936,9 @@ static int power_rate_get_sys_fs(
 	*inaccurate = true;
 
 	if ((dir = opendir(SYS_CLASS_POWER_SUPPLY)) == NULL) {
-		fprintf(stderr, "Machine does not have %s, cannot run the test.\n", SYS_CLASS_POWER_SUPPLY);
+		fprintf(stderr, "Device does not have %s, "
+			"cannot run the test.\n",
+			SYS_CLASS_POWER_SUPPLY);
 		return -1;
 	}
 
@@ -745,7 +951,8 @@ static int power_rate_get_sys_fs(
 			FILE *fp;
 
 			/* Check that type field matches the expected type */
-			snprintf(path, sizeof(path), "%s/%s/type", SYS_CLASS_POWER_SUPPLY, dirent->d_name);
+			(void)snprintf(path, sizeof(path), "%s/%s/type",
+				SYS_CLASS_POWER_SUPPLY, dirent->d_name);
 			if ((data = file_get(path)) != NULL) {
 				bool mismatch = (strstr(data, "Battery") == NULL);
 				free(data);
@@ -754,10 +961,13 @@ static int power_rate_get_sys_fs(
 			} else
 				continue;		/* can't check type, skip this entry */
 
-			snprintf(path, sizeof(path), "%s/%s/uevent", SYS_CLASS_POWER_SUPPLY, dirent->d_name);
+			(void)snprintf(path, sizeof(path), "%s/%s/uevent",
+				SYS_CLASS_POWER_SUPPLY, dirent->d_name);
 			if ((fp = fopen(path, "r")) == NULL) {
-				fprintf(stderr, "Battery %s present but under supported - no state present.", dirent->d_name);
-				closedir(dir);
+				fprintf(stderr, "Battery %s present but under "
+					"supported - no state present.",
+					dirent->d_name);
+				(void)closedir(dir);
 				return -1;
 			} else {
 				char buffer[4096];
@@ -803,7 +1013,7 @@ static int power_rate_get_sys_fs(
 				}
 				total_watts    += watts_rate + voltage * amps_rate;
 				total_capacity += watts_left + voltage * amps_left;
-				fclose(fp);
+				(void)fclose(fp);
 			}
 		}
 	} while (dirent);
@@ -811,7 +1021,7 @@ static int power_rate_get_sys_fs(
 	(void)closedir(dir);
 
 	if (! *discharging) {
-		printf("Machine is not discharging, cannot measure power usage.\n");
+		printf("Device is not discharging, cannot measure power usage.\n");
 		return -1;
 	}
 
@@ -851,7 +1061,9 @@ static int power_rate_get_proc_acpi(
 	*inaccurate = true;
 
 	if ((dir = opendir(PROC_ACPI_BATTERY)) == NULL) {
-		fprintf(stderr, "Machine does not have %s, cannot run the test.\n", PROC_ACPI_BATTERY);
+		fprintf(stderr, "Device does not have %s, "
+			"cannot run the test.\n",
+			PROC_ACPI_BATTERY);
 		return -1;
 	}
 
@@ -908,7 +1120,7 @@ static int power_rate_get_proc_acpi(
 		(void)fclose(file);
 
 		/*
-		 * Some HP firmware is broken and has an undefined 
+		 * Some HP firmware is broken and has an undefined
 		 * 'present voltage' field and instead returns this in
 		 * the design_voltage field, so work around this.
 		 */
@@ -925,7 +1137,7 @@ static int power_rate_get_proc_acpi(
 						}
 					}
 				}
-				fclose(file);
+				(void)fclose(file);
 			}
 		}
 
@@ -935,7 +1147,7 @@ static int power_rate_get_proc_acpi(
 	(void)closedir(dir);
 
 	if (! *discharging) {
-		printf("Machine is indicating it is not discharging and hence "
+		printf("Device is indicating it is not discharging and hence "
 		       "we cannot measure power usage.\n");
 		return -1;
 	}
@@ -957,6 +1169,10 @@ static int power_rate_get_proc_acpi(
 	return 0;
 }
 
+/*
+ *  power_rate_get()
+ *	fetch power via which ever interface is available
+ */
 static int power_rate_get(
 	double *const rate,
 	bool *const discharging,
@@ -972,7 +1188,7 @@ static int power_rate_get(
 	    S_ISDIR(buf.st_mode))
 		return power_rate_get_proc_acpi(rate, discharging, inaccurate);
 
-	fprintf(stderr, "Machine does not seem to have a battery, cannot measure power.\n");
+	fprintf(stderr, "Device does not seem to have a battery, cannot measure power.\n");
 	return -1;
 }
 
@@ -999,7 +1215,7 @@ static int proc_cmdline(
 	int n = 0;
 
 	*cmdline = '\0';
-	snprintf(path, sizeof(path), "/proc/%d/cmdline", pid);
+	(void)snprintf(path, sizeof(path), "/proc/%d/cmdline", pid);
 	if ((fp = fopen(path, "r")) != NULL) {
 		n = fread(cmdline, size, 1, fp);
 		(void)fclose(fp);
@@ -1083,16 +1299,16 @@ static int proc_info_add(const pid_t pid)
 	memset(cmdline, 0, sizeof(cmdline));	/* keep valgrind happy */
 
 	if ((info = calloc(1, sizeof(proc_info_t))) == NULL) {
-		fprintf(stderr, "Cannot allocate all proc info\n");
+		fprintf(stderr, "Cannot allocate all proc info.\n");
 		return -1;
 	}
 	info->pid = pid;
 
-	snprintf(path, sizeof(path), "/proc/%d/cmdline", info->pid);
+	(void)snprintf(path, sizeof(path), "/proc/%d/cmdline", info->pid);
 	(void)proc_cmdline(pid, cmdline, sizeof(cmdline));
 
 	if ((info->cmdline = malloc(strlen(cmdline)+1)) == NULL) {
-		fprintf(stderr, "Cannot allocate all proc info\n");
+		fprintf(stderr, "Cannot allocate all proc info.\n");
 		free(info);
 		return -1;
 	}
@@ -1134,11 +1350,12 @@ static int proc_info_load(void)
 static int monitor(const int sock)
 {
 	ssize_t len;
+	int64_t t = 1;
 	int redone = 0, row = 0;
 	long int readings = 0;
 	stats_t *stats, s1, s2, average, stddev, min, max;
 	struct nlmsghdr *nlmsghdr;
-	struct timeval t1, t2;
+	double time_start;
 
 	if ((stats = calloc((size_t)max_readings, sizeof(stats_t))) == NULL) {
 		fprintf(stderr, "Cannot allocate statistics table.\n");
@@ -1154,80 +1371,97 @@ static int monitor(const int sock)
 	stats_headings();
 	row++;
 
-	gettimeofday(&t1, NULL);
-	stats_read(&s1);
+	if ((time_start = gettime_to_double()) < 0.0) {
+		free(stats);
+		return -1;
+	}
+
+	if (stats_read(&s1) < 0) {
+		free(stats);
+		return -1;
+	}
 
 	while (!stop_recv && (readings < max_readings)) {
-		bool redo = false;
-		int ret;
-		suseconds_t usec;
-		struct timeval tv;
-		char __attribute__ ((aligned(NLMSG_ALIGNTO)))buf[4096];
+		double time_now, secs;
+		int ret = 0;
 
-		(void)gettimeofday(&t2, NULL);
-		usec = ((t1.tv_sec + sample_delay - t2.tv_sec) * 1000000) + (t1.tv_usec - t2.tv_usec);
-		if (usec < 0)
-			goto sample_now;
-		tv.tv_sec = usec / 1000000;
-		tv.tv_usec = usec % 1000000;
-
-		if (opts & OPTS_USE_NETLINK) {
-			fd_set readfds;
-			FD_ZERO(&readfds);
-			FD_SET(sock, &readfds);
-			ret = select(sock+1, &readfds, NULL, NULL, &tv);
-		} else {
-			ret = select(0, NULL, NULL, NULL, &tv);
-		}
-
-		if (ret < 0) {
-			if (errno == EINTR)
-				break;
-			fprintf(stderr,"select: %s\n", strerror(errno));
+		if ((time_now = gettime_to_double()) < 0.0) {
 			free(stats);
 			return -1;
 		}
+		/* Timeout to wait for in the future for this sample */
+		secs = time_start + ((double)t * sample_delay) - time_now;
+
+		if (secs > 0.0) {
+			struct timeval tv = double_to_timeval(secs);
+
+			if (opts & OPTS_USE_NETLINK) {
+				fd_set readfds;
+				FD_ZERO(&readfds);
+				FD_SET(sock, &readfds);
+				ret = select(sock+1, &readfds, NULL, NULL, &tv);
+			} else {
+				ret = select(0, NULL, NULL, NULL, &tv);
+			}
+
+			if (ret < 0) {
+				if (errno == EINTR)
+					break;
+				fprintf(stderr,"select failed: errno=%d (%s).\n",
+					errno, strerror(errno));
+				free(stats);
+				return -1;
+			}
+		}
 
 		/* Time out, so measure some more samples */
 		if (ret == 0) {
 			char tmbuffer[10];
 			bool discharging;
 
-sample_now:
 			if (redone) {
 				char buffer[80];
 				int indent;
-				snprintf(buffer, sizeof(buffer), "--- Skipped samples(s) because of %s%s%s ---",
-					redone & OPTS_REDO_WHEN_NOT_IDLE ? "low CPU idle" : "",
+				(void)snprintf(buffer, sizeof(buffer),
+					"--- Skipped samples(s) because of %s%s%s ---",
+					(redone & OPTS_REDO_WHEN_NOT_IDLE) ? "low CPU idle" : "",
 					(redone & (OPTS_REDO_WHEN_NOT_IDLE | OPTS_REDO_NETLINK_BUSY)) ==
 					(OPTS_REDO_WHEN_NOT_IDLE | OPTS_REDO_NETLINK_BUSY) ? " and " : "",
-					redone & OPTS_REDO_NETLINK_BUSY ? "fork/exec/exit activity" : "");
+					(redone & OPTS_REDO_NETLINK_BUSY) ? "fork/exec/exit activity" : "");
 				indent = (80 - strlen(buffer)) / 2;
 				row_increment(&row);
 				printf("%*.*s%s\n", indent, indent, "", buffer);
 				redone = 0;
 			}
 
-			time_now(tmbuffer, sizeof(tmbuffer));
-			(void)gettimeofday(&t1, NULL);
-			stats_read(&s2);
+			get_time(tmbuffer, sizeof(tmbuffer));
+			if (stats_read(&s2) < 0) {
+				free(stats);
+				return -1;
+			}
 
 			/*
-			 *  Total ticks was zero, something is broken, so re-sample
+			 *  Total ticks was zero, something is broken,
+			 *  so re-sample
 			 */
 			if (!stats_gather(&s1, &s2, &stats[readings])) {
 				stats_clear(&stats[readings]);
-				stats_read(&s1);
-				gettimeofday(&t1, NULL);
+				if (stats_read(&s1) < 0) {
+					free(stats);
+					return -1;
+				}
 				redone |= OPTS_REDO_WHEN_NOT_IDLE;
 				continue;
 			}
 
 			if ((opts & OPTS_REDO_WHEN_NOT_IDLE) &&
+			    (!stats[readings].inaccurate[CPU_IDLE]) &&
 			    (stats[readings].value[CPU_IDLE] < idle_threshold)) {
 				stats_clear(&stats[readings]);
-				stats_read(&s1);
-				gettimeofday(&t1, NULL);
+				if (stats_read(&s1) < 0) {
+					free(stats);
+					return -1;
+				}
 				redone |= OPTS_REDO_WHEN_NOT_IDLE;
 				continue;
 			}
@@ -1248,10 +1482,14 @@ sample_now:
 			stats_print(tmbuffer, false, &stats[readings]);
 			readings++;
 			s1 = s2;
+			t++;
 			continue;
 		}
 
 		if (opts & OPTS_USE_NETLINK) {
+			bool redo = false;
+			char __attribute__ ((aligned(NLMSG_ALIGNTO)))buf[4096];
+
         		if ((len = recv(sock, buf, sizeof(buf), 0)) == 0) {
 				free(stats);
 				return 0;
@@ -1260,13 +1498,17 @@ sample_now:
 				if (errno == EINTR) {
 					continue;
 				} else {
-					fprintf(stderr,"recv: %s\n", strerror(errno));
+					fprintf(stderr,"recv failed: errno=%d (%s).\n",
+						errno, strerror(errno));
 					free(stats);
             				return -1;
 				}
 			}
 
-			for (nlmsghdr = (struct nlmsghdr *)buf; NLMSG_OK (nlmsghdr, len); nlmsghdr = NLMSG_NEXT (nlmsghdr, len)) {
+			for (nlmsghdr = (struct nlmsghdr *)buf;
+				NLMSG_OK (nlmsghdr, len);
+				nlmsghdr = NLMSG_NEXT (nlmsghdr, len)) {
+
 				struct cn_msg *cn_msg;
 				struct proc_event *proc_ev;
 
@@ -1328,15 +1570,21 @@ sample_now:
 			/* Have we been asked to redo a sample on fork/exec/exit? */
 			if (opts & OPTS_REDO_NETLINK_BUSY && redo) {
 				stats_clear(&stats[readings]);
-				stats_read(&s1);
-				gettimeofday(&t1, NULL);
+				if (stats_read(&s1) < 0) {
+					free(stats);
+					return -1;
+				}
 				redone |= OPTS_REDO_NETLINK_BUSY;
 			}
         	}
 	}
 
-	/* Stats now gathered, calculate averages, stddev, min and max and display */
-	stats_average_stddev_min_max(stats, readings, &average, &stddev, &min, &max);
+	/*
+	 * Stats now gathered, calculate averages, stddev,
+	 * min and max and display
+	 */
+	stats_average_stddev_min_max(stats, readings, &average,
+		&stddev, &min, &max);
 	if (readings > 0) {
 		stats_ruler();
 		stats_print("Average", true, &average);
@@ -1369,7 +1617,7 @@ sample_now:
  */
 void show_help(char *const argv[])
 {
-	printf("%s, version %s\n\n", APP_NAME, VERSION);
+	printf("%s, version %s\n\n", app_name, VERSION);
 	printf("usage: %s [-d secs] [-i idle] [-b|-h|-p|-r|-s|-z] [delay [count]]\n", argv[0]);
 	printf("\t-b redo a sample if a system is busy, considered less than %d%% CPU idle\n", IDLE_THRESHOLD);
 	printf("\t-d specify delay before starting, default is %ld seconds\n", start_delay);
@@ -1380,7 +1628,7 @@ void show_help(char *const argv[])
 	printf("\t-s show process fork/exec/exit activity log\n");
 	printf("\t-S calculate power from capacity drain using standard average\n");
 	printf("\t-z forcibly ignore zero power rate stats from the battery\n");
-	printf("\tdelay: delay between each sample, default is %d seconds\n", SAMPLE_DELAY);
+	printf("\tdelay: delay between each sample, default is %.1f seconds\n", SAMPLE_DELAY);
 	printf("\tcount: number of samples to take\n");
 }
 
@@ -1390,9 +1638,7 @@ int main(int argc, char * const argv[])
 	int sock = -1, ret = EXIT_FAILURE, i;
 	long int run_duration;
 	bool discharging, dummy_inaccurate;
-
-    	signal(SIGINT, &handle_sigint);
-    	siginterrupt(SIGINT, 1);
+	struct sigaction new_action;
 
 	for (;;) {
 		int c = getopt(argc, argv, "bd:hi:prszS");
@@ -1406,11 +1652,11 @@ int main(int argc, char * const argv[])
 			errno = 0;
 			start_delay = strtol(optarg, NULL, 10);
 			if (errno) {
-				fprintf(stderr, "Invalid value for start delay\n");
+				fprintf(stderr, "Invalid value for start delay.\n");
 				exit(EXIT_FAILURE);
 			}
 			if (start_delay < 0) {
-				fprintf(stderr, "Start delay must be 0 or more seconds\n");
+				fprintf(stderr, "Start delay must be 0 or more seconds.\n");
 				exit(EXIT_FAILURE);
 			}
 			break;
@@ -1421,7 +1667,7 @@ int main(int argc, char * const argv[])
 			opts |= OPTS_REDO_WHEN_NOT_IDLE;
 			idle_threshold = atof(optarg);
 			if ((idle_threshold < 0.0) || (idle_threshold > 99.99)) {
-				fprintf(stderr, "Idle threshold must be between 0..99.99\n");
+				fprintf(stderr, "Idle threshold must be between 0..99.99.\n");
 				exit(EXIT_FAILURE);
 			}
 			break;
@@ -1440,6 +1686,9 @@ int main(int argc, char * const argv[])
 		case 'z':
 			opts |= OPTS_ZERO_RATE_ALLOW;
 			break;
+		case '?':
+			printf("Try '%s -h' for more information.\n", app_name);
+			exit(EXIT_FAILURE);
 		default:
 			show_help(argv);
 			exit(EXIT_FAILURE);
@@ -1448,13 +1697,13 @@ int main(int argc, char * const argv[])
 
 	if (optind < argc) {
 		errno = 0;
-		sample_delay = strtol(argv[optind++], NULL, 10);
+		sample_delay = atof(argv[optind++]);
 		if (errno) {
-			fprintf(stderr, "Invalid value for start delay\n");
+			fprintf(stderr, "Invalid value for start delay.\n");
 			exit(EXIT_FAILURE);
 		}
-		if (sample_delay < 1) {
-			fprintf(stderr, "Sample delay must be >= 1\n");
+		if (sample_delay < 0.5) {
+			fprintf(stderr, "Sample delay must be greater or equal to 0.5 seconds.\n");
 			exit(EXIT_FAILURE);
 		}
 	}
@@ -1465,37 +1714,44 @@ int main(int argc, char * const argv[])
 		errno = 0;
 		max_readings = strtol(argv[optind++], NULL, 10);
 		if (errno) {
-			fprintf(stderr, "Invalid value for maximum readings\n");
+			fprintf(stderr, "Invalid value for maximum readings.\n");
 			exit(EXIT_FAILURE);
 		}
 		if ((max_readings * sample_delay) < run_duration) {
-			fprintf(stderr, "Number of readings should be at least %ld\n",
-				run_duration / sample_delay);
+			fprintf(stderr, "Number of readings should be at least %ld.\n",
+				(long int)(run_duration / sample_delay));
 			exit(EXIT_FAILURE);
 		}
 	} else {
 		max_readings = run_duration / sample_delay;
 	}
 
+	if (max_readings < 5) {
+		fprintf(stderr, "Number of readings too low.\n");
+		exit(EXIT_FAILURE);
+	}
+	if (max_readings < 10)
+		fprintf(stderr, "Number of readings low, results may be inaccurate.\n");
+
 	if (geteuid() == 0)
 		opts |= OPTS_ROOT_PRIV;
 	else if (opts & OPTS_USE_NETLINK) {
-		fprintf(stderr, "%s needs to be run with root privilege when using -p, -r, -s options\n", argv[0]);
+		fprintf(stderr, "%s needs to be run with root privilege when using -p, -r, -s options.\n", argv[0]);
 		exit(ret);
 	}
 
 	if (power_rate_get(&dummy_rate, &discharging, &dummy_inaccurate) < 0)
 		exit(ret);
-	printf("Running for %ld seconds (%ld samples at %ld second intervals).\n",
+	printf("Running for %.1f seconds (%ld samples at %.1f second intervals).\n",
 			sample_delay * max_readings, max_readings, sample_delay);
-	printf("ACPI battery power measurements will start in %ld seconds time\n",
+	printf("Battery power measurements will start in %ld seconds time.\n",
 		start_delay);
 	printf("\n");
 
 	if (start_delay > 0) {
 		/* Gather up initial data */
 		for (i = 0; i < start_delay; i++) {
-			printf("Waiting %ld seconds before starting (gathering samples) \r", start_delay - i);
+			printf("Waiting %ld seconds before starting (gathering samples). \r", start_delay - i);
 			fflush(stdout);
 			if (power_rate_get(&dummy_rate, &discharging, &dummy_inaccurate) < 0)
 				exit(ret);
@@ -1507,6 +1763,21 @@ int main(int argc, char * const argv[])
 		printf("%79.79s\r", "");
 	}
 
+	memset(&new_action, 0, sizeof(new_action));
+	for (i = 0; signals[i] != -1; i++) {
+		new_action.sa_handler = handle_sig;
+		sigemptyset(&new_action.sa_mask);
+		new_action.sa_flags = 0;
+
+		if (sigaction(signals[i], &new_action, NULL) < 0) {
+			fprintf(stderr, "sigaction failed: errno=%d (%s).\n",
+				errno, strerror(errno));
+			exit(EXIT_FAILURE);
+		}
+		(void)siginterrupt(signals[i], 1);
+	}
+
+
 	log_init();
 	if (opts & OPTS_USE_NETLINK) {
     		sock = netlink_connect();
@@ -1537,8 +1808,8 @@ abort:
 	if (opts & OPTS_USE_NETLINK) {
 		log_dump();
 		log_free();
-		if (sock != -1)
-			close(sock);
+		if (sock >= 0)
+			(void)close(sock);
 	}
 
 	exit(ret);
